/*
* Copyright 2009, KyTea Development Team
* 
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
* 
*     http://www.apache.org/licenses/LICENSE-2.0
* 
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#ifndef GENERAL_IO_H__ 
#define GENERAL_IO_H__ 

#include "kytea-string.h"
#include "string-util.h"
#include <iostream>
#include <fstream>
#include <sstream>

// do sanity checks on the IO files
#define KYTEA_IO_SAFE

using namespace std;

namespace kytea {

class GeneralIO {

protected:
    
    StringUtil* util_;
    iostream* str_;
    bool out_;
    bool bin_;
    bool owns_;
 
    // write
    template <class T>
    inline void writeBinary(T v) {
        str_->write(reinterpret_cast<char *>(&v), sizeof(T));
    } 
    inline void writeString(const char* str, size_t size) {
        str_->write(str, size+1);
    } 
    inline void writeString(const string & str) {
        str_->write(str.c_str(), str.length()+1);
    }
    inline void writeString(const KyteaString & str) {
        writeBinary(str.length());
        for(unsigned i = 0; i < str.length(); i++)
            writeBinary(str[i]);
    }
    
    // read
    template <class T>
    inline T readBinary() {
        T v;
        str_->read(reinterpret_cast<char *>(&v),sizeof(T));
        return v;
    } 
    inline string readString() {
        string str;
        getline(*str_, str, (char)0);
        return str;
    }
    inline KyteaString readKyteaString() {
        KyteaString ret(readBinary<unsigned>());
        for(unsigned i = 0; i < ret.length(); i++)
            ret[i] = readBinary<KyteaChar>();
        return ret;
    }
    

public:

    GeneralIO(StringUtil* util) : 
        util_(util), str_(0), out_(true), 
        bin_(false), owns_(false) { }

    GeneralIO(StringUtil* util, iostream & str, bool out, bool bin) : 
        util_(util), str_(&str), out_(out), bin_(false), owns_(false) 
            { setStream(str, out, bin); }

    GeneralIO(StringUtil* util, const char* file, bool out, bool bin) : 
        util_(util), str_(0), bin_(false), owns_(true) 
            { openFile(file,out,bin); }
    
    ~GeneralIO() {
        if(str_ && owns_)
            delete str_;
    }

    void openFile(const char* file, bool out, bool bin) {
        fstream::openmode mode = (out?fstream::out:fstream::in);
        if(bin) out = out | fstream::binary;
        fstream * str = new fstream(file, mode);
        if(str->fail()) {
            delete str;
            ostringstream buff;
            buff << "Couldn't open file '"<<file<<"' for "<<(out?"output":"input");
            throw runtime_error(buff.str());
        }
        setStream(*str, out, bin);
        owns_ = true;
    }

    void setStream(iostream & str, bool out, bool bin) {
        if(str_ && owns_)
            delete str_;
        str_ = &str;
        bin_ = bin;
        out_ = out;
        owns_ = false;
    }

};

}

#endif
